/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package actor;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;
import logging.Logger;
import org.w3c.dom.Node;

/**
 * Oberklasse fuer eine Actorkomponente.
 * @author nico3000
 */
public abstract class ActorComponent {
    /**
     * Alle bekannten Komponenten Packages.
     */
    private static final Set<String> componentPackages = new LinkedHashSet<>();
    
    /**
     * Besitzender Actor.
     */
    private Actor owner;
    
    /**
     * Liefert die eindeutige ID der gesamten Komponentenklasse.
     * @return Die ID
     */
    protected abstract int getComponentId();
    
    /**
     * Wird zur Initialisierung der Komponente aufgerufen.
     * @param data Daten fuer die Initialisierung
     * @return <tt>true</tt>, falls erfolgreich
     */
    protected abstract boolean init(Node data);
    
    /**
     * Wird direkt nach erfolgreicher Initialisierung aufgerufen.
     */
    protected void postInit() {}
    
    /**
     * Wird jeden Frame aufgerufen.
     * @param deltaMillis Die Anzahl der Millisekunden (Gametimer) des letzten
     * Frames
     */
    protected void update(long deltaMillis) {}
    
    /**
     * Zerstoert die Komponente.
     */
    protected void destroy() {}
    
    /**
     * Setzt den besitzenden Actor.
     * @param owner Der besitzende Actor
     */
    final void setOwner(Actor owner) {
        this.owner = owner;
    }
    
    /**
     * Getter fuer den besitzenden Actor.
     * @return Der besitzende Actor
     */
    public final Actor getOwner() {
        return this.owner;
    }
    
    public static boolean addComponentPackage(String pakkage) {
        if(pakkage != null) {
            if(!pakkage.isEmpty() && !pakkage.endsWith(".")) {
                pakkage += ".";
            }
            return componentPackages.add(pakkage);
        } else {
            return false;
        }
    }
    
    /**
     * Erzeugt eine leere Komponente anhand des Namens.
     * @param name Name der Komponente
     * @return Die leere Komponente oder <tt>null</tt>, bei ungueltigem Namen
     */
    static ActorComponent createComponent(String name) {
        ActorComponent component = null;
        Iterator<String> iter = componentPackages.iterator();
        while(iter.hasNext() && component == null) {
            Class compClass = null;
            try {
                compClass = Class.forName(iter.next() + name);
            } catch (ClassNotFoundException ex) {
                // wrong package, do nothing
            }
            if(compClass != null) {
                Object instance = null;
                try {
                    instance = compClass.newInstance();
                } catch (InstantiationException | IllegalAccessException ex) {
                    // some serious error
                    Logger.INSTANCE.error(ex.getMessage());
                }
                if(instance instanceof ActorComponent) {
                    component = (ActorComponent)instance;
                }
            }
        }
        if(component == null) {
            Logger.INSTANCE.error("\"" + name + "\" is no known ActorComponent!");
        }
        return component;
    }
}
