/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package event;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import logging.Logger;
import gameapp.GameApp;

/**
 * Sammelt Instanzen von <tt>Event</tt> solange, bis sie mit
 * <tt>processEvents()</tt> an die registrierten Instanzen von Klassen, die das
 * Interface <tt>EventListener</tt> implementieren, verteilt werden.
 * @author nico3000
 */
public class EventManager {
    public static final int TYPE_ALL = 0xffffffff;
    
    private List<EventData> currentEvents = new LinkedList<>();
    private List<EventData> nextEvents = new LinkedList<>();
    private final Map<Integer, List<EventListener>> listeners = new HashMap<>();
    
    /**
     * Verteilt alle seit dem letzten Aufruf gesammelten Events an die Listener.
     */
    public void processEvents() {
        List<EventData> toProcess = this.currentEvents;
        this.currentEvents = this.nextEvents;
        
        for(EventData event : toProcess) {
            this.triggerEvent(event);
        }
        toProcess.clear();
        
        this.nextEvents = toProcess;
    }
    
    /**
     * Reiht ein Event an das Ende der bisher gesammelten ein.
     * @param event Das neue Event
     */
    public void queueEvent(EventData event) {
        event.timeStamp = GameApp.getRealTimer().getTimeStamp();
        this.currentEvents.add(event);
    }
    
    /**
     * Verteilt das Event sofort an alle Listener.
     * @param event Das Event
     */
    public void triggerEvent(EventData event) {
        List<EventListener> list = this.listeners.get(event.getId());
        if(list != null) {
            for(EventListener listener : list) {
                listener.trigger(event);
            }
        }
    }
    
    /**
     * Registriert einen Listener fuer bestimmte Eventtypen.
     * @param listener Der zu registrierende Listener
     * @param types Die Eventtypen, auf die der Listener hoeren soll
     */
    public boolean register(EventListener listener, int... types) {
        boolean success = true;
        for(int type : types) {
            List<EventListener> list = this.listeners.get(type);
            if(list == null) {
                this.listeners.put(type, list = new LinkedList<>());
            }
            if(list.contains(listener)) {
                Logger.INSTANCE.warning("Tried to register listener twice.");
                success = false;
            } else {
                list.add(listener);
            }
        }
        return success;
    }
    
    /**
     * Entfernt einen Listener fuer bestimmte Eventtypen.
     * @param listener Der Listener
     * @param types Die Eventtypen, auf die der Listener nicht mehr hoeren soll
     */
    public boolean unregister(EventListener listener, int... types) {
        boolean success = false;
        if(types.length == 1 && types[0] == TYPE_ALL) {
            for(List<EventListener> list : this.listeners.values()) {
                if(list.remove(listener)) {
                    success = true;
                }
            }
            if(!success) {
                Logger.INSTANCE.warning("Tried to remove unknown listener.");
            }
        } else {
            success = true;
            for(int type : types) {
                List<EventListener> list = this.listeners.get(type);
                if(list != null) {
                    if(!list.remove(listener)) {
                        Logger.INSTANCE.warning("Tried to remove listener from non listening type (" + listener + ").");
                        success = false;
                    }
                }
            }
        }
        return success;
    }
}
