/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package gameapp;

import java.util.LinkedList;
import java.util.List;

import javax.vecmath.Vector3f;

import logging.Logger;

import actor.Actor;

import com.bulletphysics.collision.broadphase.AxisSweep3_32;
import com.bulletphysics.collision.broadphase.BroadphaseInterface;
import com.bulletphysics.collision.dispatch.CollisionConfiguration;
import com.bulletphysics.collision.dispatch.CollisionDispatcher;
import com.bulletphysics.collision.dispatch.CollisionObject;
import com.bulletphysics.collision.dispatch.DefaultCollisionConfiguration;
import com.bulletphysics.collision.dispatch.GhostPairCallback;
import com.bulletphysics.collision.narrowphase.PersistentManifold;
import com.bulletphysics.dynamics.DiscreteDynamicsWorld;
import com.bulletphysics.dynamics.DynamicsWorld;
import com.bulletphysics.dynamics.RigidBody;
import com.bulletphysics.dynamics.constraintsolver.ConstraintSolver;
import com.bulletphysics.dynamics.constraintsolver.SequentialImpulseConstraintSolver;

/**
 *
 * @author nico3000
 */
public class PhysicsLayer {
    /**
     * Timestep in Millisekunden.
     */
    private static final long DESIRED_TIME_STEP_MILLIS = 10;
    
    /**
     * Akkumulator der vergangenen Millisekunden.
     */
    private long milliAccu = 0;
    /**
     * JBullet CollisionConfiguration.
     */
    private CollisionConfiguration collConf = null;
    /**
     * JBullet CollisionDispatcher.
     */
    private CollisionDispatcher collDisp = null;
    /**
     * JBullet ConstraintSolver.
     */
    private ConstraintSolver solver = null;
    /**
     * JBullet DynamicsWorld.
     */
    private DynamicsWorld world = null;
    private final List<RigidBody> bodies = new LinkedList<>();
    private BroadphaseInterface overlappingPairCache = new AxisSweep3_32(new Vector3f(-100.0f, -100.0f, -100.0f), new Vector3f(100.0f, 100.0f, 100.0f));
    
    /**
     * Initialisiert das PhysicsLayer.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        this.collConf = new DefaultCollisionConfiguration();
        this.collDisp = new CollisionDispatcher(this.collConf);
        this.solver = new SequentialImpulseConstraintSolver();
        this.world = new DiscreteDynamicsWorld(this.collDisp, this.overlappingPairCache, this.solver, this.collConf);
        this.overlappingPairCache.getOverlappingPairCache().setInternalGhostPairCallback(new GhostPairCallback());
        
        this.world.setGravity(new Vector3f(0.0f, -9.81f, 0.0f));
        
        return true;
    }
    
    /**
     * Fuehrt einen Simulationsschritt durch, wenn insgesamt genug Zeit
     * vergangen ist.
     * @param deltaMillis Dauer des Frames in Millisekunden (Gametimer)
     */
    public void update(long deltaMillis) {
        //this.world.stepSimulation(1e-3f * (float)deltaMillis);
        if((this.milliAccu += deltaMillis) >= DESIRED_TIME_STEP_MILLIS) {
            this.world.stepSimulation(1e-3f * (float)this.milliAccu);
            this.milliAccu = 0;
        }
    }
    
    public void addRigidBodyToWorld(RigidBody body) {
        this.world.addRigidBody(body);
        this.bodies.add(body);
    }
    
    public void addCollisionObjectToWorld(CollisionObject collObj) {
    	this.world.addCollisionObject(collObj);
    }
    
    public void removeRigidBodyFromWorld(RigidBody toRemove) {
        this.bodies.remove(toRemove);
        this.world.removeRigidBody(toRemove);
        for(RigidBody body : this.bodies) {
            body.activate();
        }
    }

    public DynamicsWorld getWorld() {
        return world;
    }
    
    /**
     * Zerstoert das PhysicsLayer.
     */
    public void destroy() {
        
    }
}
