/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh.loader.wavefront;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;
import logging.Logger;
import gameapp.GameApp;
import mesh.material.Material;
import mesh.material.MaterialLibrary;
import org.lwjgl.util.vector.Vector4f;
import resource.ResHandle;
import resource.Resource;

/**
 * Erzeugt aus den Daten einer Wavefront <tt>.mtl</tt> Datei eine
 * MaterialLibrary.
 * @author nico3000
 */
public class WavefrontMtlLoader {
    /**
     * Die mtl Resource.
     */
    private final String resource;
    /**
     * Aktuelles Material waehrend des Parsens.
     */
    private Material currentMaterial = null;
    /**
     * Aktuelle Zeile des Parsens.
     */
    private int currentLine;
    /**
     * Die Bibliothek.
     */
    private final MaterialLibrary library;
    
    /**
     * Konstruktor.
     * @param resource Der Name der Resource
     */
    public WavefrontMtlLoader(String resource) {
        this.resource = resource;
        this.library = MaterialLibrary.createLibrary(resource);
    }
    
    /**
     * Erzeugt eine MaterialLibrary.
     * @return Die erzeugte MaterialLibrary oder <tt>null</tt>, falls ein Fehler
     *         aufgetreten ist
     */
    public MaterialLibrary load() {
        if(this.library.isClosed()) {
            return this.library;
        }
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(this.resource));
        if(handle != null) {
            boolean success = false;
            BufferedReader reader = new BufferedReader(new InputStreamReader(handle.getBufferAsInputStream()));
            try {
                success = this.loadData(reader);
                reader.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            handle.unlock();
            if(success) {
                return this.library;
            }
        }
        return null;
    }
    
    /**
     * Laedt die Daten und pflegt sie in die Datenstruktur ein.
     * @param reader Die Daten
     * @return <tt>true</tt>, falls erfolgreich
     * @throws IOException Fehler des BufferedReader werden weitergereicht
     */
    private boolean loadData(BufferedReader reader) throws IOException {
        this.currentLine = 0;
        String line;
        this.currentMaterial = null;
        while((line = reader.readLine()) != null) {
            ++this.currentLine;
            line = line.trim();
            if(!line.isEmpty() && !line.startsWith("#")) {
                boolean success = true;
                StringTokenizer tokenizer = new StringTokenizer(line);
                String type = tokenizer.nextToken();
                if(this.currentMaterial == null && !type.equals("newmtl")) {
                    this.error("no current material");
                }
                String textureName = null;
                if(type.startsWith("map_")) {
                    String split[] = line.split(" ", 2);
                    if(split.length == 2) {
                        textureName = split[1];
                    } else {
                        this.error("illegal texture definition");
                        success = false;
                    }
                }
                switch(type) {
                    case "newmtl":
                        success = this.createMaterial(tokenizer);
                        break;
                    case "Ka":
                        success = this.setColor(this.currentMaterial.getKa(), tokenizer);
                        break;
                    case "Kd":
                        success = this.setColor(this.currentMaterial.getKd(), tokenizer);
                        break;
                    case "Ks":
                        success = this.setColor(this.currentMaterial.getKs(), tokenizer);
                        break;
                    case "map_d":
                        this.currentMaterial.setAlphaTextue(textureName);
                        break;
                    case "map_Ka":
                        this.currentMaterial.setAmbientTextue(textureName);
                        break;
                    case "map_Kd":                        
                        this.currentMaterial.setDiffuseTextue(textureName);
                        break;
                    case "map_Ks":                        
                        this.currentMaterial.setSpecularTextue(textureName);
                        break;
                    default:
                        //Logger.INSTANCE.warning(String.format("unsupported or illegal type \"%s\" was ignored (%s, %d)", type, this.resource, this.currentLine));
                        break;
                }
                if(!success) {
                    return false;
                }
            }
        }
        this.finishCurrentMaterial();
        this.library.setClosed();
        return true;
    }

    /**
     * Erzeugt ein neues Material.
     * @param data Die Daten des Materials
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean createMaterial(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.error("Illegal newmtl line!");
            return false;
        }
        this.finishCurrentMaterial();
        String name = data.nextToken();
        this.currentMaterial = this.library.createMaterial(name);
        return this.currentMaterial != null;
    }
    
    /**
     * Liest eine Farbe aus einem String.
     * @param dest Der Zielvektor
     * @param data Die Daten der Farbe
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setColor(Vector4f dest, StringTokenizer data) {
        if(data.countTokens() != 3 && data.countTokens() != 4) {
            this.error("Wrong data count!");
        }
        try {
            dest.x = Float.parseFloat(data.nextToken());
            dest.y = Float.parseFloat(data.nextToken());
            dest.z = Float.parseFloat(data.nextToken());
            if (data.hasMoreTokens()) {
                dest.w = Float.parseFloat(data.nextToken());
            } else {
                dest.w = 1.0f;
            }
        } catch (NumberFormatException ex) {
            this.error("Illegal float expression!");
            return false;            
        }
        return true;
    }
    
    private void finishCurrentMaterial() {
        if(this.currentMaterial != null) {
            this.currentMaterial.updateData();
            this.currentMaterial.release();
            this.currentMaterial = null;
        }
    }
    
    
    /**
     * Loggt einen Fehler und fuegt den Namen und die Zeile der obj-Datei an.
     * @param message Die Fehlermeldung
     */
    private void error(String message) {
        Logger.INSTANCE.error(1, "%s (%s, %d)", message, this.resource, this.currentLine);
    }
}
