/*
 * This software is the proprietary information of nico3000.
 */
package mesh.material;

import java.util.HashMap;
import java.util.Map;
import logging.Logger;

/**
 * Stellt eine Bibliothek von Materialien zur Verfuegung. Ein Material kann
 * mittels des Namens gefordert werden.
 * @author nico3000
 */
public class MaterialLibrary {
    /**
     * Map aller Instanzen der MaterialLibrary sortiert nach ihren Namen.
     */
    private static final Map<String, MaterialLibrary> INSTANCES = new HashMap<>();
    
    /**
     * Liefert eine MaterialLibrary mit einem bestimmten Namen und erstellt sie,
     * falls sie nicht vorhanden ist. Der Name darf dabei nicht leer sein.
     * @param name Der Name der MaterialLibrary
     * @return Die MaterialLibrary oder <tt>null</tt>, falls nicht erfolgreich
     */
    public static MaterialLibrary createLibrary(String name) {
        if(name == null || (name = name.trim()).isEmpty()) {
            Logger.INSTANCE.error("MaterialLibrary with empty name!");
            return null;
        } else {
            MaterialLibrary library = INSTANCES.get(name);
            if(library == null) {
                //Logger.INSTANCE.info("Now creating MaterialLibrary %s.", name);
                INSTANCES.put(name, library = new MaterialLibrary(name));
            }
            ++library.ref;
            return library;
        }        
    }
    
    /**
     * Liefert eine MaterialLibrary anhand ihres Namens, falls sie existiert.
     * @param name Der Name der MaterialLibrary
     * @return Die MaterialLibrary, falls sie existiert; sonst <tt>null</tt>
     */
    public static MaterialLibrary getLibrary(String name) {
        MaterialLibrary library = INSTANCES.get(name);
        if(library != null) {
            ++library.ref;
        }
        return library;
    }
    
    /**
     * Gibt den Status der bekannten MaterialLibrarys zurueck.
     * @return Der Status
     */
    public static String getDebugStatus() {
        String librarys = "";
        for(MaterialLibrary library : INSTANCES.values()) {
            if(library.isClosed()) {
                librarys += String.format("\n- %s (%d)", library.name, library.ref);
            }
        }
        return String.format("Active MaterialLibrary instances: %s", librarys.isEmpty() ? "none" : librarys);
    }
    
    /**
     * Der Name der Bibliothek.
     */
    private final String name;
    /**
     * Die Materialien nach Namen sortiert.
     */
    private final Map<String, Material> materials = new HashMap<>();
    /**
     * Gibt an, ob die MaterialLibrary abgeschlossen ist. Einer abgeschlossenen
     * Library koennen keine weiteren Materialien hinzugefuegt werden.
     */
    private boolean closed = false;
    /**
     * Der Reference Counter dieser MaterialLibrary.
     */
    private int ref = 0;

    /**
     * Konstruktor.
     * @param name Der Name der Bibliothek
     */
    private MaterialLibrary(String name) {
        this.name = name == null ? "Generic library" : name;
    }

    /**
     * Liefert den Namen der MaterialLibrary-
     * @return Der Name
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * Liefert ein Material mit einem bestimmten Namen. Der Reference Counter
     * des Materials wird automatisch um 1 erhoeht. Daher muss auf dem Material
     * <tt>release()</tt> aufgerufen werden, sobald es nicht mehr benoetigt
     * wird.
     * @param name Der Name des geforderten Materials
     * @return Das Material, oder <tt>null</tt>, falls es keine mit dem
     *         gegebenen Namen gibt
     */
    public Material getMaterialByName(String name) {
        Material material = this.materials.get(name);
        if(material != null) {
            material.addRef();
        }
        return material;
    }
    
    /**
     * Erstellt ein neues Material in dieser Bibliothek mit einem Namen, falls 
     * diese nicht bereits abgeschlossen ist. Wird das Material nicht mehr
     * benutzt, muss es mit <tt>release()</tt> freigegeben werden. Der Name darf
     * nicht leer sein und es darf kein Material mit demselben Namen bereits in
     * der Bibliothek vorhanden sein.
     * @param material Das Material
     * @return Das Material, falls erfolgreich; sonst <tt>null</tt>
     */
    public Material createMaterial(String materialName) {
        if(this.closed) {
            Logger.INSTANCE.error("Trying to create new material in already closed library %s!", this.name);
            return null;
        } else if(materialName == null || (materialName = materialName.trim()).isEmpty()) {
            Logger.INSTANCE.error("Trying toregister material with no name in library %s!", this.name);
            return null;
        } else if(this.materials.containsKey(materialName)) {
            Logger.INSTANCE.error("A material of the same name (%s) already exists in library (%s)!", materialName, this.name);
            return null;
        } else {
            Material material = new Material(this.name, materialName);
            if(!material.init()) {
                Logger.INSTANCE.error("Material %s in MaterialLibrary %s could not be initialized!", materialName, this.name);
                return null;
            }
            this.materials.put(materialName, material);
            material.addRef();
            return material;
        }
    }
    
    /**
     * Schliesst die MaterialLibrary ab.
     */
    public void setClosed() {
        this.closed = true;
    }

    /**
     * Liefert, ob die MaterialLibrary bereits abgeschlossen ist.
     * @return <tt>true</tt>, falls abgeschlossen
     */
    public boolean isClosed() {
        return this.closed;
    }
    
    /**
     * Gibt eine Referenz der MaterialLibrary wieder frei. Wurden alle
     * Referenzen freigegeben, werden automatisch alle Materialien frei
     * gegegeben. In diesem Fall wird fuer jedes weiterhin referenzierte
     * Material eine Warning ausgegeben.
     */
    public void release() {
        if(this.ref == 0) {
            //Logger.INSTANCE.error("Cannot release MaterialLibrary %s because reference counter is already 0!", this.name);
        } else if(--this.ref == 0) {
            //Logger.INSTANCE.info("Now releasing MaterialLibrary %s.", this.name);
            for(Material material : this.materials.values()) {
                if(material.getReferenceCount() != 0) {
                	//TODO: Warning gets called
                    //Logger.INSTANCE.warning("Material %s in released material library %s still referenced somewhere (%d).", material.getName(), this.name, material.getReferenceCount());
                }
                material.destroy();
            }
            this.materials.clear();
            this.closed = false;
        }
    }
}
