/*
 * This software is the proprietary information of nico3000.
 */
package n3gui;

import graphics.texture.Texture;
import java.awt.Rectangle;
import logging.Logger;

/**
 *
 * @author nico3000
 */
public abstract class Component {
    /**
     * Neu zeichnen, wenn keine Textur gesetzt ist.
     */
    public static final int REPAINT_ON_NULL_TEXTURE = 1;
    /**
     * Neu zeichnen, wenn sich die Komponente bewegt hat.
     */
    public static final int REPAINT_ON_MOVE = 2;
    /**
     * Neu zeichnen, wenn sich die Groesse der Komponente geaendert hat.
     */
    public static final int REPAINT_ON_RESIZE = 4;
    /**
     * Neu zeichnen, wenn die Komponente den Fokus erhalten oder verloren hat.
     */
    public static final int REPAINT_ON_FOCUS_CHANGED = 8;
    /**
     * Neu zeichnen, wenn der Cursor die Komponente betreten oder verlassen hat.
     */
    public static final int REPAINT_ON_HOVER_CHANGED = 16;
    /**
     * In allen Faellen neu zeichnen.
     */
    public static final int REPAINT_ALL = ~0;
    
    /**
     * Die Textur der Komponente.
     */
    private Texture texture = null;
    /**
     * Der zugehorige Knoten.
     */
    private GuiTreeNode node = null;
    /**
     * Die Bedingungen des Neuzeichnens.
     */
    private int repaintOn = 0;
    /**
     * Gibt an, ob die Komponente neu gezeichnet werden muss.
     */
    private boolean mustRepaint = false;
    /**
     * Gibt an, ob die Komponente auf Fokusevents reagiert.
     */
    private boolean focusable = true;
    /**
     * Gibt an, ob die Komponente auf Hoverevents reagiert.
     */
    private boolean hoverable = true;
    /**
     * Das Clippingrechteck im lokalen Koordinatensystem.
     */
    private final int contentPadding[] = { 0, 0, 0, 0, };
    /**
     * Gibt den Offset des Vaterknotens an, der das Element clippen kann.
     */
    private int clipOffset = 0;
    
    /**
     * Setzt Bits, die bestimmen, wann die Komponente sich neu zeichnen muss.
     * @param bits Die Bits, die gesetzt werden sollen
     */
    protected void setRepaintBits(int bits) {
        this.repaintOn |= bits;
    }
    
    /**
     * Loescht Bits, die bestimmen, wann die Komponente sich neu zeichnen muss.
     * @param bits Die Bits, die geloescht werden soll
     */
    protected void clearRepaintBits(int bits) {
        this.repaintOn &= ~bits;
    }

    /**
     * Gibt an, ob die Komponente auf Fokusevents reagiert.
     * @return <tt>true</tt>, falls die Komponente auf Fokusevents reagiert
     */
    public boolean isFocusable() {
        return focusable;
    }

    /**
     * Setzt, ob die Komponente auf Fokusevents reagiert.
     * @param focusable <tt>true</tt>, falls die Komponente auf Fokusevents
     *                  reagiert
     */
    public void setFocusable(boolean focusable) {
        this.focusable = focusable;
    }
    
    /**
     * Gibt an, ob die Komponente auf Hoverevents reagiert.
     * @return <tt>true</tt>, falls die Komponente auf Hoverevents reagiert
     */
    public boolean isHoverable() {
        return hoverable;
    }
    
    /**
     * Setzt, ob die Komponente auf Hoverevents reagiert.
     * @return <tt>true</tt>, falls die Komponente auf Hoverevents reagiert
     */
    public void setHoverable(boolean hoverable) {
        this.hoverable = hoverable;
    }
    
    /**
     * Makriert die Komponente als neu zu zeichnen, falls ein Repaintbit gesetzt
     * ist.
     * @param bit Das Bit
     */
    void repaintIfBitIsSet(int bit) {
        if((this.repaintOn & bit) != 0) {
            this.mustRepaint = true;
        }
    }
    
    /**
     * Markiert die Komponente als neu zu zeichnend.
     */
    protected void mustRepaint() {
        this.mustRepaint = true;
    }

    int getClipOffset() {
        return this.clipOffset;
    }

    public void setClipOffset(int clipOffset) {
        this.clipOffset = clipOffset;
    }
    
    /**
     * Setzt eine neue Textur. Die alte wird dabei freigegeben.
     * @param texture Die neue Textur
     */
    protected void setTexture(Texture texture) {
        if(this.texture != null) {
            this.texture.release();
        }
        this.texture = texture;
        if(this.texture != null) {
            this.texture.addRef();
        }
    }
    
    /**
     * Zerstoert die Komponente.
     */
    public void destroy() {
        this.setTexture(null);
    }

    /**
     * Setzt den Knoten dieser Komponente. Ist das Setzen einmal erfolgreich,
     * darf die Methode nicht erneut aufgerufen werden.
     * @param node Der Node
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean setNode(GuiTreeNode node) {
        if(node == null) {
            Logger.INSTANCE.warning("Component does not allow null node.");
            return false;
        } else if(this.node != null) {
            Logger.INSTANCE.warning("Component node already set (%s).", this.node);
            return false;
        } else {
            this.node = node;
            return true;
        }
    }
    
    /**
     * Liefert die zugehoerige Node.
     * @return Die Node
     */
    public GuiTreeNode getNode() {
        return this.node;
    }
    
    /**
     * Zeichnet die Komponente neu, falls bestimmte Bedingungen erfuellt sind
     * und bindet die Textur an einen bestimmten Slot.
     * @param slot Der Slot
     * @return <tt>true</tt>, falls die Textur gebunden wurde
     */
    public boolean bindTexture(int slot) {
        if(this.texture == null) {
            this.repaintIfBitIsSet(REPAINT_ON_NULL_TEXTURE);
        }
        if(this.node.getDimension().hasPosChanged()) {
            this.repaintIfBitIsSet(REPAINT_ON_MOVE);
        }
        if(this.node.getDimension().hasSizeChanged()) {
            this.repaintIfBitIsSet(REPAINT_ON_RESIZE);
        }
        if(this.mustRepaint) {
            this.repaint();
            this.mustRepaint = false;
        }
        if(this.texture != null) {
            this.texture.connectToSlot(slot);
            this.texture.bind();
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Liefert die bevorzugte Breite der Komponente.
     * @return Die bevorzugte Breite
     */
    protected int getPreferredWidth() {
        return 0;
    }
    
    /**
     * Liefert die bevorzugte Hoehe der Komponente.
     * @return Die bevorzugte Hoehe
     */
    protected int getPreferredHeight() {
        return 0;
    }
    
    /**
     * Wird aufgerufen, sobald die Komponente einem Knoten zugewiesen und dieser
     * komplett initialisiert wurde.
     */
    protected void onInitialized() {}
    
    /**
     * Wird aufgerufen, wenn die Textur neu generiert werden soll.
     */
    protected abstract void repaint();
    
    /**
     * Wird aufgerufen, wenn sich der Cursor der Gui bewegt hat. Kann von einer
     * Unterklasse ueberschrieben werden.
     * @param cursorX Die aktuelle absolute x-Koordinate des Cursors
     * @param cursorY Die aktuelle absolute y-Koordinate des Cursors
     * @param dX Die x-Koordinate des Bewegungsvektors
     * @param dY Die y-Koordinate des Bewegungsvektors
     */
    protected void onCursorMoved(int cursorX, int cursorY, int dX, int dY) {}
    
    /**
     * Wird aufgerufen, wenn die Komponente fokussiert werden kann und der
     * Button des Pointers ueber der Komponente gedrueckt wurde.
     * @param button Der Button
     */
    protected void onButtonDown(int button) {}
    
    /**
     * Wird aufgerufen, wenn die Komponente fokussiert ist und der Button des
     * Pointers losgelassen wurde.
     * @param button Der Button
     */
    protected void onButtonUp(int button) {}
    
    /**
     * Wird aufgerufen, sobald sich die Groesse des Elements aendert.
     */
    protected void onSizeChanged() {}
    
    /**
     * Wird aufgerufen, sobald das Element bewegt wurde.
     */
    protected void onMoved() {}
    
    /**
     * Wird aufgerufen, sobald der Cursor ueber das Element bewegt wird oder das
     * Element verlaesst.
     */
    protected void onHoverChanged() {}
    
    /**
     * Wird aufgerufen, sobald das Element den Fokus erhaelt oder verliert.
     */
    protected void onFocusChanged() {}
    
    /**
     * Wird jeden Frame vor dem Rendern aufgerufen.
     * @param deltaMillis Die Dauer des letzten Frame sin Millisekunden
     */
    protected void onUpdate(long deltaMillis) {}
    
    /**
     * Prueft, ob der Cursor in einem gegebenen Rechteck enthalten ist. Der
     * Nullpunkt des Koordinatensystem des Rechtecks ist dabei die linke obere
     * Ecke der Komponente.
     * @param relRect Das Rechteck
     * @return <tt>true</tt>, falls der Cursor in dem Rechteck ist
     */
    protected boolean containsCursor(Rectangle relRect) {
        int relX = this.getNode().getTree().getCursorX() - this.getNode().getDimension().getGlobalX();
        int relY = this.getNode().getTree().getCursorY() - this.getNode().getDimension().getGlobalY();
        return relRect.contains(relX, relY);
    }
    
    /**
     * Liefert ein Rechteck, dessen Inhalt den Kindkomponenten zum Zeichnen zur
     * Verfuegung steht.
     * @return Das Clippingrechteck
     */
    protected Rectangle getClippingRect() {
        return new Rectangle(
                this.getNode().getDimension().getGlobalX() + this.contentPadding[0],
                this.getNode().getDimension().getGlobalY() + this.contentPadding[1],
                this.getNode().getDimension().getWidth() - this.contentPadding[0] - this.contentPadding[2],
                this.getNode().getDimension().getHeight() - this.contentPadding[1] - this.contentPadding[3]);
    }
    
    /**
     * Setzt den Abstand vom Rand der Komponente zum Inhalt.
     * @param left Der Abstand zum linken Rand
     * @param top Der Abstand zu oberen Rand
     * @param right Der Abstand zum rechten Rand
     * @param bottom Der Abstand zum unteren Rand
     */
    protected void setContentPadding(int left, int top, int right, int bottom) {
        this.contentPadding[0] = left;
        this.contentPadding[1] = top;
        this.contentPadding[2] = right;
        this.contentPadding[3] = bottom;
        if(this.node != null) {
            this.node.getDimension().sizeChanged();
        }
    }

    @Override
    public String toString() {
        return this.getNode().toString();
    }
}
