/*
 * This software is the proprietary information of nico3000.
 */
package n3gui;

import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;
import graphics.shader.Effect;
import input.PointerHandler;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import logging.Logger;
import mesh.Mesh;
import mesh.loader.generic.GenericLoader;
import n3gui.components.Cursor;
import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;

/**
 *
 * @author nico3000
 */
public class GuiTree implements PointerHandler {
    /**
     * Ein Screenquad.
     */
    public static Mesh QUAD;
    /**
     * Der Effekt, um die Gui zu zeichnen.
     */
    public static final Effect GUI_EFFECT = new Effect();
    /**
     * Der Texturslot der Guitextur eines Elements.
     */
    public static final int EFFECT_TEX_SLOT = 7;
    
    /**
     * Initialisiert das Quad und den Effekt der GUI.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public static boolean initGui() {
        if(!GUI_EFFECT.initVF("Shaders/N3GuiVS.glsl", "Shaders/N3GuiFS.glsl")) {
            Logger.INSTANCE.error("Initialization of n3 gui effect failed!");
            return false;
        }
        GUI_EFFECT.use();
        GUI_EFFECT.setUniform("g_ComponentTex", EFFECT_TEX_SLOT);
        IndexBuffer iBuffer = new IndexBuffer();
        VertexBuffer vBuffer = new VertexBuffer();
        if(!iBuffer.init() || !vBuffer.init()) {
            Logger.INSTANCE.error("Initialization of n3 gui quad geometry failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();
            return false;
        }
        IntBuffer indices = BufferUtils.createIntBuffer(4);
        indices.put(new int[] { 0, 1, 3, 2, });
        indices.position(0);
        iBuffer.setData(indices, IndexBuffer.TRIANGLE_STRIP);
        FloatBuffer vertices = BufferUtils.createFloatBuffer(8);
        vertices.put(new float[] { 0.0f, 0.0f, 
                                   1.0f, 0.0f, 
                                   1.0f, 1.0f, 
                                   0.0f, 1.0f, });
        vertices.position(0);
        vBuffer.setData(vertices);
        vBuffer.addAttribute(VertexBuffer.ATTR_POS, 2);
        QUAD = Mesh.createMesh(new GenericLoader(vBuffer, iBuffer));
        if(QUAD == null) {
            Logger.INSTANCE.error("Initialization of n3 gui quad failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();
            return false;
        }
        return true;
    }
    
    /**
     * Zerstoert das Quad und den Effekt der GUI.
     */
    public static void destroyGui() {
        QUAD.release();
        GUI_EFFECT.destroy();
    }    
    
    /**
     * Alle Nodes des Baums nach ihrer ID.
     */
    private final Map<String,GuiTreeNode> nodes = new HashMap<>();
    /**
     * Der Wurzelknoten.
     */
    private final GuiTreeNode root = new GuiTreeNode("root", null, this);
    /**
     * Der Cursor der Gui.
     */
    private final Cursor cursor = new Cursor();
    /**
     * Das erste Element unter dem Cursor.
     */
    private GuiTreeNode hovered = null;
    /**
     * Das Element mit dem primaeren Fokus.
     */
    private GuiTreeNode focused = null;
    /**
     * Der DrawHelper dieser Gui.
     */
    private final DrawHelper helper = new DrawHelper(null);
    /**
     * Das i-te Bit gibt an, ob Button i des Pointers gedrueckt ist.
     */
    private int buttonStates = 0;

    /**
     * Konstruktor.
     * @param guiWidth Die Breite der Gui in Pixel
     * @param guiHeight Die Hoehe der Gui in Pixel
     */
    public GuiTree(int guiWidth, int guiHeight) {
        this.root.getDimension().setSize(guiWidth, guiHeight);
    }
    
    /**
     * Initialisiert die Gui.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        if(!this.root.init()) {
            Logger.INSTANCE.warning("Failed to initialize root node of %s.", this);
            return false;
        } else if(this.addComponent(null, "cursor", this.cursor) == null) {
            Logger.INSTANCE.warning("Failed to add cursor to %s.", this);
            this.destroy();
            return false;
        } else {
            return true;
        }
    }
    
    /**
     * Zerstoert die komplette Gui.
     */
    public void destroy() {
        this.root.destroy();
    }
    
    /**
     * Liefert die globale x-Koordinate des Cursors.
     * @return Die x-Koordinate des Cursors
     */
    public int getCursorX() {
        return this.cursor.getCursorX();
    }
    
    /**
     * Liefert die globale y-Koordinate des Cursors.
     * @return Die y-Koordinate des Cursors
     */
    public int getCursorY() {
        return this.cursor.getCursorY();
    }

    /**
     * Liefert den DrawHelper dieser Gui.
     * @return Der DrawHelper
     */
    public DrawHelper getDrawHelper() {
        return helper;
    }
    
    /**
     * Liefert einen Knoten anhand seines Qualifiers.
     * @param qualifier Der Qualifier
     * @return Der Knoten oder <tt>null</tt>, falls der Qualifier keinen
     *         existierenden Knoten beschreibt
     */
    protected GuiTreeNode findByQualifier(String qualifier) {
        if(qualifier == null) {
            return this.root;
        } else {
            GuiTreeNode node = this.root;
            int index = 0;
            for(String name : qualifier.split("\\.")) {
                node = node.findChildByName(name);
                if(node == null) {
                    Logger.INSTANCE.warning("Qualifier %s broken at index %d (%s).", qualifier, index, name);
                    return null;
                }
            }
            return node;
        }
    }
    
    /**
     * Liefert einen Knoten anhand seiner ID.
     * @param id Die ID des Knotens
     * @return Der Knoten, falls vorhanden
     */
    protected GuiTreeNode getById(String id) {
        return this.nodes.get(id);
    }
    
    /**
     * Fuegt dem Baum eine Komponente hinzu.
     * @param parentQualifier Der Vater, oder <tt>null</tt>, falls keiner
     *                        vorhanden
     * @param id Die ID der Komponente
     * @param component Die Komponente
     * @return Die Dimension des Knotens, falls erfolgreich; sonst <tt>null</tt>
     */
    public Dimension addComponent(String parentQualifier, String id, Component component) {
        GuiTreeNode parent = this.findByQualifier(parentQualifier);
        if(parent == null) {
            Logger.INSTANCE.warning("Invalid qualifier %s.", parentQualifier);
            return null;
        }
        GuiTreeNode child = new GuiTreeNode(id, component, this);
        if(component != null && !component.setNode(child)) {
            Logger.INSTANCE.warning("Failed to link node %s with its component.", child);
            return null;
        } else if(!child.init()) {
            Logger.INSTANCE.warning("Failed to initialize %s.", child);
            return null;
        } else if(parent.addChild(child)) {
            this.nodes.put(id, child);
            component.onInitialized();
            return child.getDimension();
        } else {
            Logger.INSTANCE.warning("Error while adding %s to tree.", child);
            return null;
        }
    }
    
    /**
     * Zeichnet verschiedenfarbige Umrisse fuer jedes Element.
     */
    public void drawSchemes() {
        GL11.glDisable(GL11.GL_DEPTH_TEST);
        GL11.glDisable(GL11.GL_CULL_FACE);
        GUI_EFFECT.use();
        this.root.drawSchemes();
        GL11.glEnable(GL11.GL_CULL_FACE);
        GL11.glEnable(GL11.GL_DEPTH_TEST);
    }
    
    /**
     * Liefert eine iterierbare Postantiorder Traversierung des Baums. Dabei
     * werden von einem Knoten zuerst die Kinder in umgekehrter Reihenfolge
     * durchlaufen und anschließend der Knoten selbst.
     * @return Die Traversierung
     */
    public Iterable<GuiTreeNode> postAntiOrder() {
        return new Iterable<GuiTreeNode>() {
            @Override
            public Iterator<GuiTreeNode> iterator() {
                return new PostAntiOrderIterator();
            }
        };
    }
    
    /**
     * Liefert eine iterierbare Preorder Traversierung des Baums.
     * @return Die Traversierung
     */
    public Iterable<GuiTreeNode> preOrder() {
        return new Iterable<GuiTreeNode>() {
            @Override
            public Iterator<GuiTreeNode> iterator() {
                return new PreOrderIterator();
            }
        };
    }
    
    /**
     * Updatet alle Knoten des Baums.
     * @param deltaMillis Die Dauer des letzten Frames in Millieksunden
     */
    public void update(long deltaMillis) {
        this.root.update(deltaMillis);
    }

    @Override
    public boolean pointerMoved(int posX, int posY, int dX, int dY) {
        // move cursor
        int oldCursorX = this.cursor.getCursorX();
        int oldCursorY = this.cursor.getCursorY();
        this.cursor.moveCursor(dX, -dY);
        int newCursorX = this.cursor.getCursorX();
        int newCursorY = this.cursor.getCursorY();
        int cursorDX = newCursorX - oldCursorX;
        int cursorDY = newCursorY - oldCursorY;
        if(this.hovered != null) {
            this.hovered.setHovered(false);
        }
        // propagate moved cursor
        if(cursorDX != 0 || cursorDY != 0) {
            boolean updateFocused = this.focused != null;
            boolean consumed = false;
            Iterator<GuiTreeNode> iter = this.postAntiOrder().iterator();
            while(!consumed && iter.hasNext()) {
                GuiTreeNode node = iter.next();
                if(node.contains(newCursorX, newCursorY) && node.setHovered(true)) {
                    consumed = true;
                    this.hovered = node;
                    node.getComponent().onCursorMoved(newCursorX, newCursorY, cursorDX, cursorDY);
                    updateFocused &= node != this.focused;
                }
            }
            if(updateFocused) {
                this.focused.getComponent().onCursorMoved(newCursorX, newCursorX, cursorDX, cursorDY);
            }
        }
        return this.focused != null;
    }

    @Override
    public boolean onButtonDown(int button) {
        this.buttonStates |= 1 << button;
        int cursorX = this.cursor.getCursorX();
        int cursorY = this.cursor.getCursorY();
        Iterator<GuiTreeNode> iter = this.postAntiOrder().iterator();
        boolean consumed = false;
        if(this.focused != null) {
            this.focused.setFocused(false);
            this.focused = null;
        }
        //Logger.INSTANCE.info("Checking focus.");
        while(!consumed && iter.hasNext()) {
            GuiTreeNode node = iter.next();
            if(node.contains(cursorX, cursorY)) {
                consumed = node.setFocused(true);
                if(consumed) {
                    this.focused = node;
                    this.focused.getComponent().onButtonDown(button);
                    //Logger.INSTANCE.info("%s did accept the focus.", node);
                } else {
                    //Logger.INSTANCE.info("%s didn't accept the focus.", node);
                }
            } else {
                //Logger.INSTANCE.info("Not in %s.", node);
            }
        }
        return consumed;
    }

    @Override
    public boolean onButtonUp(int button) {
        this.buttonStates &= ~(1 << button);
        if(this.focused != null) {
            this.focused.getComponent().onButtonUp(button);
            return true;
        } else {
            return false;
        }        
    }

    @Override
    public boolean onWheelMoved(int delta) {
        return false;
    }
    
    /**
     * Iterator, der eine Postantiorder Traversierung des Baums realisiert.
     */
    private class PostAntiOrderIterator implements Iterator<GuiTreeNode> {
        /**
         * Hilfsstack.
         */
        private final Stack<GuiTreeNode> stack = new Stack<>();
        /**
         * Markieren der Knoten, deren Kinder bereits auf dem Stack liegen /
         * lagen.
         */
        private final Set<GuiTreeNode> marked = new HashSet<>();

        /**
         * Konstruktor
         */
        public PostAntiOrderIterator() {
            this.stack.push(GuiTree.this.root);
        }

        @Override
        public boolean hasNext() {
            return !this.stack.isEmpty();
        }

        @Override
        public GuiTreeNode next() {
            GuiTreeNode node = this.stack.pop();
            while(!this.marked.contains(node)) {
                this.marked.add(node);
                this.stack.push(node);
                for(GuiTreeNode child : node.getChildren()) {
                    this.stack.push(child);
                }
                node = this.stack.pop();
            }
            return node;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Removing of nodes not supported.");
        }
    }
    
    /**
     * Iterator, der eine Preorder Traversieurng des Baums realisiert.
     */
    private class PreOrderIterator implements Iterator<GuiTreeNode> {
        /**
         * Hilfsstack.
         */
        private final Stack<GuiTreeNode> stack = new Stack<>();
        
        /**
         * Konstruktor.
         */
        public PreOrderIterator() {
            this.stack.push(GuiTree.this.root);
        }
        
        @Override
        public boolean hasNext() {
            return this.stack.isEmpty();
        }

        @Override
        public GuiTreeNode next() {
            GuiTreeNode next = this.stack.pop();
            ListIterator<GuiTreeNode> iter = next.getChildren().listIterator(next.getChildren().size());
            while(iter.hasPrevious()) {
                this.stack.push(iter.previous());
            }
            return next;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Removing of nodes not supported.");
        }
        
    }
}
