/*
 * This software is the proprietary information of nico3000.
 */
package n3gui.components;

import graphics.texture.Texture;
import graphics.texture.TextureLoader;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import n3gui.Component;
import n3gui.DrawHelper;
import n3gui.GuiAction;

/**
 *
 * @author nico3000
 */
public class Button extends Component {
    public static enum ButtonColor {
        BLUE,
        GREEN,
        RED,
        PURPLE,
        ORANGE,
        GRAY,
        YELLOW,
    }
    
    private String text;
    private ButtonColor color = ButtonColor.BLUE;
    private Texture defaultTex;
    private Texture defaultHoveredTex;
    private Texture pressedTex;
    private GuiAction action;
    private boolean pressed;
    private boolean completeRepaint = true;
    
    
    public Button(String text, ButtonColor color) {
        this.setRepaintBits(REPAINT_ON_NULL_TEXTURE | REPAINT_ON_HOVER_CHANGED);
        this.text = text;
        this.color = color;
    }
    
    public void setText(String text) {
        this.text = text;
        this.completeRepaint = true;
        this.mustRepaint();
    }

    public void setColor(ButtonColor color) {
        this.color = color;
        this.completeRepaint = true;
        this.mustRepaint();
    }

    public void setAction(GuiAction action) {
        this.action = action;
    }

    @Override
    public void destroy() {
        this.pressedTex.release();
        this.defaultTex.release();
        this.defaultHoveredTex.release();
        super.destroy();
    }

    @Override
    protected void onButtonDown(int button) {
        if(button == 0) {
            this.pressed = true;
            this.mustRepaint();
        }
    }

    @Override
    protected void onButtonUp(int button) {
        if(button == 0) {
            this.pressed = false;
            if(this.action != null && this.getNode().isHovered()) {
                this.action.fire();
            }
            this.mustRepaint();
        }
    }
    
    @Override
    protected void repaint() {
        if(this.completeRepaint) {
            this.completeRepaint();
        }
        if(this.getNode().isHovered()) {
            this.setTexture(this.pressed ? this.pressedTex : this.defaultHoveredTex);
        } else {
            this.setTexture(this.defaultTex);
        }
    }
    
    private void completeRepaint() {
        if(this.defaultTex != null) {
            this.defaultTex.release();
        }
        if(this.pressedTex != null) {
            this.pressedTex.release();
        }
        if(this.defaultHoveredTex != null) {
            this.defaultHoveredTex.release();
        }
        switch(this.color) {
            case BLUE:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundBlue.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundBlueBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundBlueInv.png", true));
                break;
            case GREEN:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGreen.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGreenBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGreenInv.png", true));
                break;
            case RED:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundRed.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundRedBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundRedInv.png", true));
                break;
            case PURPLE:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundPurple.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundPurpleBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundPurpleInv.png", true));
                break;
            case ORANGE:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundOrange.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundOrangeBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundOrangeInv.png", true));
                break;
            case GRAY:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGray.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGrayBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundGrayInv.png", true));
                break;
            case YELLOW:
                this.defaultTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundYellow.png", false));
                this.defaultHoveredTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundYellowBrighter.png", false));
                this.pressedTex = TextureLoader.createTexture(this.createButton("Textures/N3Gui/BackgroundYellowInv.png", true));
                break;
        }
        this.completeRepaint = false;
    } 
    
    private BufferedImage createButton(String backgroundResource, boolean pressed) {
        int width = this.getNode().getDimension().getWidth();
        int height = this.getNode().getDimension().getHeight();
        int radius = Math.min(12, Math.min(width / 2, height / 2));
        DrawHelper helper = this.getNode().getTree().getDrawHelper();
        BufferedImage background = helper.getImage(backgroundResource);
        BufferedImage image = helper.createBaseImage(width, height);
        Graphics2D graphics = image.createGraphics();
        graphics.setColor(helper.getDefaultColor(DrawHelper.ColorType.BORDER));
        graphics.setClip(new RoundRectangle2D.Float(0, 0, width - 1, height - 1, radius, radius));
        graphics.drawImage(background, 0, 0, width, height, null);
        graphics.setClip(null);
        graphics.drawRoundRect(0, 0, width - 1, height - 1, radius, radius);            
        if(this.text != null) {
            if(this.text.startsWith("res:")) {
                BufferedImage buttonImage = helper.getImage(this.text.substring(4));
                graphics.drawImage(buttonImage, 6, 6, width - 12, height - 12, null);
            } else {
                graphics.setColor(helper.getDefaultColor(DrawHelper.ColorType.FONT));
                graphics.setFont(helper.getDefaultFont());
                helper.drawText(this.text, graphics, new Rectangle(4, 4, width - 8, height - 8), DrawHelper.ALIGN_H_C | DrawHelper.ALIGN_V_C, pressed ? 2 : 1);
            }
        }
        return image;
    }
}
