/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package resource.file;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import logging.Logger;
import resource.Resource;

/**
 * Eine ResourceFile, die Dateien aus einer zip-Datei bezieht.
 * @author nico3000
 */
public class ZipResourceFile implements ResourceFile {
    /**
     * Der Dateiname der Zipdatei.
     */
    private final String filename;
    /**
     * Die Zipdatei.
     */
    private ZipFile file;
    /**
     * Lowercase Dateinamen werden auf die vorhandenen gemapt.
     */
    private Map<String, String> entries = new HashMap<>();

    /**
     * Konstruktor.
     * @param filename Name der zip-Datei
     */
    public ZipResourceFile(String filename) {
        this.filename = filename;
        this.file = null;
    }

    @Override
    public boolean open() {
        try {
            this.file = new ZipFile(this.filename);
        } catch (IOException ex) {
            Logger.INSTANCE.error(ex.getMessage());
            return false;
        }
        Enumeration<? extends ZipEntry> enumeration = this.file.entries();
        while(enumeration.hasMoreElements()) {
            ZipEntry entry = enumeration.nextElement();
            this.entries.put(entry.getName().toLowerCase(), entry.getName());
        }
        return true;
    }

    @Override
    public void close() {
        if(this.file != null) {
            try {
                this.file.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
        }
    }

    @Override
    public int getRawResourceSize(Resource resource) {
        ZipEntry entry = this.getEntryByResourceName(resource);
        if(entry == null) {
            Logger.INSTANCE.error("unknown resource: " + resource.getName());
            return 0;
        }
        return (int)entry.getSize();
    }

    @Override
    public int getRawResource(Resource resource, byte buffer[]) {
        int read = 0;
        InputStream stream = null;
        ZipEntry entry = this.getEntryByResourceName(resource);
        if(entry == null) {
            Logger.INSTANCE.error("unknown resource: " + resource.getName());
            return 0;
        }
        try {
            stream = this.file.getInputStream(entry);
            while(read != buffer.length) {
                read += stream.read(buffer, read, buffer.length - read);        // TODO: may block forever?
            }
        } catch (IOException ex) {
            Logger.INSTANCE.error(ex.getMessage());
        } finally {
            if(stream != null) {
                try {
                    stream.close();
                } catch (IOException ex) {
                    Logger.INSTANCE.error(ex.getMessage());
                }
            }
        }
        return read;
    }

    @Override
    public int getNumResources() {
        Enumeration<? extends ZipEntry> enumeration = this.file.entries();
        int count = 0;
        while(enumeration.hasMoreElements()) {
            ++count;
            enumeration.nextElement();
        }
        return count;
    }

    @Override
    public String getResourceName(int num) {
        Enumeration<? extends ZipEntry> enumeration = this.file.entries();
        while(enumeration.hasMoreElements()) {
            if(num-- == 0) {
                return enumeration.nextElement().getName();
            }
        }
        return null;
    }   
    
    /**
     * Liefert zu einer Resource den zugehoerigen ZipEntry.
     * @param resource Die Resource
     * @return Der zugehoerige Eintrag der zip-Datei
     */
    private ZipEntry getEntryByResourceName(Resource resource) {
        return this.file.getEntry(this.entries.get(resource.getName().toLowerCase()));
    }
}
