/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package util;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import logging.Logger;

import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Hilfsmethoden zu XML Nodes.
 * @author nico3000
 */
public class XMLUtil {
    /**
     * Interpretiert den Wert eines Attributes als Integer.
     * @param node Der Node dessen Attribut genutzt werden soll
     * @param attr Das Attribut
     * @param defaultValue Der Standardwert, falls das Parsen fehlschlaegt
     * @return Der Wert des Attributes als Integer
     */
    public static int getAttributeAsInt(Node node, String attr, int defaultValue) {
        Node attrNode = node == null ? null : node.getAttributes().getNamedItem(attr);
        int returnValue = defaultValue;
        if(attrNode != null) {
            try {
                returnValue = Integer.parseInt(attrNode.getTextContent());
            } catch (NumberFormatException ex) {
                // nothing to do...
            }
        }
        return returnValue;
    }
    
    /**
     * Interpretiert den Wert eines Attributes als Boolean.
     * @param node Der Node dessen Attribut genutzt werden soll
     * @param attr Das Attribut
     * @param defaultValue Der Standardwert, falls das Parsen fehlschlaegt
     * @return Der Wert des Attributes als Boolean
     */
    public static boolean getAttributeAsBoolean(Node node, String attr, boolean defaultValue) {
        Node attrNode = node == null ? null : node.getAttributes().getNamedItem(attr);
        boolean returnValue = defaultValue;
        if(attrNode != null) {
            try {
                returnValue = Boolean.parseBoolean(attrNode.getTextContent());
            } catch (NumberFormatException ex) {
                // nothing to do...
            }
        }
        return returnValue;
    }
    
    /**
     * Interpretiert den Wert eines Attributes als Float.
     * @param node Der Node dessen Attribut genutzt werden soll
     * @param attr Das Attribut
     * @param defaultValue Der Standardwert, falls das Parsen fehlschlaegt
     * @return Der Wert des Attributes als Float
     */
    public static float getAttributeAsFloat(Node node, String attr, float defaultValue) {
        Node attrNode = node == null ? null : node.getAttributes().getNamedItem(attr);
        float returnValue = defaultValue;
        if(attrNode != null) {
            try {
                returnValue = Float.parseFloat(attrNode.getTextContent());
            } catch (NumberFormatException ex) {
                // nothing to do...
            }
        }
        return returnValue;
    }
    
    public static Vector3f getAttributesAsVector(Node node, String attrX, String attrY, String attrZ, Vector3f dest) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.x = getAttributeAsFloat(node, attrX, dest.x);
        dest.y = getAttributeAsFloat(node, attrY, dest.y);
        dest.z = getAttributeAsFloat(node, attrZ, dest.z);
        return dest;
    }
    
    public static Vector2f getAttributesAsVector(Node node, String attrX, String attrY, Vector2f dest) {
        if(dest == null) {
            dest = new Vector2f();
        }
        dest.x = getAttributeAsFloat(node, attrX, dest.x);
        dest.y = getAttributeAsFloat(node, attrY, dest.y);
        return dest;
    }
    
    /**
     * Sucht einen direkten Nachfahren eines Nodes anhand des Typs.
     * @param parent Der Knoten, dessen Kinder durchsucht werden sollen
     * @param type Der Typ des zu suchenden Knotens
     * @return Der Kindknoten vom Typ <tt>type</tt> oder <tt>null</tt>, falls
     *         nicht vorhanden
     */
    public static Node getChildByType(Node parent, String type) {
        if(parent != null) {
            NodeList children = parent.getChildNodes();
            for(int i=0; i < children.getLength(); ++i) {
                if(children.item(i).getNodeName().equals(type)) {
                    return children.item(i);
                }
            }
        }
        return null;
    }
    
    public static String getAttributeAsString(Node node, String attr) {
        if(node != null) {
            node = node.getAttributes().getNamedItem(attr);
            return node == null ? null : node.getNodeValue();
        } else {
            return null;
        }
    }
    
	/**
	 * Hilfsfunktion, um ein beliebiges Document-Object als XML formatiert in einen
	 * beliebigen OutputStream zu schreiben
	 * @param doc Document-Object der XML
	 * @param out OutputStream fuer das XML-Ergebnis
	 * @throws IOException
	 * @throws TransformerException
	 */
	public static void printDocument(Document doc, OutputStream out) throws IOException, TransformerException {
        TransformerFactory tf = TransformerFactory.newInstance();
        Transformer transformer = tf.newTransformer();
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
        transformer.setOutputProperty(OutputKeys.METHOD, "xml");
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");

        transformer.transform(new DOMSource(doc), 
             new StreamResult(new OutputStreamWriter(out, "UTF-8")));
    }
}
