/*
 * This software is the proprietary information of nico3000.
 */
package view;

import controls.PointerMovedEvent;
import controls.PointerWheelEvent;
import event.events.ToggleEvent;
import event.EventData;
import input.InputHandler;
import input.KeyboardHandler;
import input.PointerHandler;
import logging.Logger;
import gameapp.GameApp;
import util.Joiner;

/**
 *
 * @author nico3000
 */
public class PlayerControls implements KeyboardHandler, PointerHandler {
    private final Joiner<Integer, Class<EventData>> keyMappings = new Joiner<>();
    private final Joiner<Integer, Class<EventData>> btnMappings = new Joiner<>();
    private Class<EventData> pointerMovedEvt;
    private Class<EventData> pointerWheelEvt;
    
    private boolean processRepeatEvents = false;
    
    public boolean init() {
        return this.initFromConfigSection("CONTROLS");
    }
    
    private boolean initFromConfigSection(String sectionName) {
    	String repeatEvents = GameApp.getConfig().getAsString("bProcessingRepeatEvents");
    	if (repeatEvents != null) {
    		this.processRepeatEvents = GameApp.getConfig().getAsBoolean("bProcessingRepeatEvents");
    	}
    	
    	
        for(String evtName : GameApp.getConfig().getSection(sectionName)) {
            String inputName = GameApp.getConfig().getAsString(evtName);
            Class<EventData> evtClass = null;
            try {
                evtClass = (Class<EventData>)Class.forName(evtName);
            } catch (ClassNotFoundException | ClassCastException ex) {
                Logger.INSTANCE.warning("Unknwon event: " + evtName + " (" + ex.getMessage() + ").");
            }
            if(evtClass != null) {
                int key = GameApp.getInputHandler().getKey(inputName);
                int button = GameApp.getInputHandler().getButton(inputName);
                if(key != InputHandler.NO_KEY) {
                    this.assignKeyEvt(key, evtClass);
                } else if(button != InputHandler.NO_BUTTON) {
                    this.assignPointerButtonEvt(button, evtClass);
                } else if(inputName.equals("MOUSE")) {
                    this.assignPointerMovedEvent(evtClass);
                } else if(inputName.equals("WHEEL")) {
                    this.assignPointerWheelEvent(evtClass);
                } else {
                    Logger.INSTANCE.warning("Unknown input name \"" + inputName + "\" for event \"" + evtName + "\".");
                }
            }
        }
        return true;
    }
    
    public void assignKeyEvt(int key, Class<EventData> evtClass) {
        this.keyMappings.join(key, evtClass);
    }
    
    public void assignPointerButtonEvt(int btn, Class<EventData> evtClass) {
        this.btnMappings.join(btn, evtClass);
    }

    public void assignPointerMovedEvent(Class<EventData> evtClass) {
        this.pointerMovedEvt = evtClass;
    }
    
    public void assignPointerWheelEvent(Class<EventData> evtClass) {
        this.pointerWheelEvt = evtClass;
    }
    
    @Override
    public boolean onKeyDown(int key, char c) {
        Class<EventData> evtClass = this.keyMappings.getJoined1(key);
        return evtClass != null && this.fireInputEvent(evtClass, true);
    }

    @Override
    public boolean onKeyUp(int key) {
        Class<EventData> evtClass = this.keyMappings.getJoined1(key);
        return evtClass != null && this.fireInputEvent(evtClass, false);
    }

    @Override
    public boolean isProcessingRepeatEvents() {
        return this.processRepeatEvents;
    }

    @Override
    public boolean pointerMoved(int posX, int posY, int dX, int dY) {
        if(this.pointerMovedEvt != null) {
            try {
                EventData evt = this.pointerMovedEvt.newInstance();
                if(evt instanceof PointerMovedEvent) {
                    ((PointerMovedEvent)evt).setDelta(dX, dY);
                }
                GameApp.getEventManager().queueEvent(evt);
            } catch (InstantiationException | IllegalAccessException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    public boolean onButtonDown(int button) {
        Class<EventData> evtClass = this.btnMappings.getJoined1(button);
        return evtClass != null && this.fireInputEvent(evtClass, true);
    }

    @Override
    public boolean onButtonUp(int button) {
        Class<EventData> evtClass = this.btnMappings.getJoined1(button);
        return evtClass != null && this.fireInputEvent(evtClass, false);
    }

    @Override
    public boolean onWheelMoved(int delta) {
        if(this.pointerWheelEvt != null) {
            try {
                EventData evt = this.pointerWheelEvt.newInstance();
                if(evt instanceof PointerWheelEvent) {
                    ((PointerWheelEvent)evt).setDelta(delta);
                }
                GameApp.getEventManager().queueEvent(evt);
            } catch (InstantiationException | IllegalAccessException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            return true;
        } else {
            return false;
        }
    }
    
    private boolean fireInputEvent(Class<EventData> evtClass, boolean inputDown) {
        EventData evt;
        try {
            evt = evtClass.newInstance();
        } catch (InstantiationException | IllegalAccessException ex) {
            Logger.INSTANCE.error(ex.getMessage());
            return false;
        }
        boolean isToggleEvent = evt instanceof ToggleEvent;
        if(isToggleEvent) {
            ((ToggleEvent)evt).setActivated(inputDown);
        }
        if(isToggleEvent || inputDown) {
            GameApp.getEventManager().queueEvent(evt);
            return true;
        } else {
            return false;
        }
    }
}
