/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package view.scene;

import java.util.LinkedList;
import logging.Logger;
import org.lwjgl.util.vector.Matrix4f;
import util.MathUtil;

/**
 * Der Matrixstack sammelt und verknuepft Transformationsmatrizen, um die
 * Traversierung des Szenegraphen zu erleichtern.
 * @author nico3000
 */
public class MatrixStack {
    /**
     * Aktueller Stack von Transformationsmatrizen.
     */
    private final LinkedList<Matrix4f> matStack = new LinkedList<>();
    /**
     * Aktueller Stack von inversen Transformationsmatrizen.
     */
    private final LinkedList<Matrix4f> matInverseStack = new LinkedList<>();
    /**
     * Die Position des aktuellen Topelements.
     */
    private int current;
    
    /**
     * Konstruktor fuer einen Matrixstack.
     */
    public MatrixStack() {
        this.matStack.add(Matrix4f.setIdentity(new Matrix4f()));
        this.matInverseStack.add(Matrix4f.setIdentity(new Matrix4f()));
        this.current = 0;
    }
    
    /**
     * Multipliziert eine Matrix mit dem aktuellen Topelement und legt das
     * Ergebnis auf den Stack.
     * @param mat Die Transformationsmatrix
     * @param matInverse Die inverse Transformationsmatrix
     */
    public void push(Matrix4f mat, Matrix4f matInverse) {
        Matrix4f destMat, destMatInverse;
        if(++this.current < this.matStack.size()) {
            destMat = this.matStack.get(this.current);
            destMatInverse = this.matInverseStack.get(this.current);
        } else {
            this.matStack.add(destMat = new Matrix4f());
            this.matInverseStack.add(destMatInverse = new Matrix4f());
        }
        MathUtil.mul(destMat, this.matStack.get(this.current - 1), mat);
        MathUtil.mul(destMatInverse, this.matInverseStack.get(this.current - 1), matInverse);
    }
    
    /**
     * Loescht das aktuelle Topelement, sofern vorhanden. <i>Hinweis:</i> Ein
     * leerer Matrixstack besitzt immer noch die Einheitsmatrix als Topelement.
     * Dieses kann durch <tt>pop()</tt> nicht entfernt werden.
     */
    public void pop() {
        if(this.current == 0) {
            Logger.INSTANCE.error("tried to pop empty stack");
        } else {
            --this.current;
        }
    }
    
    /**
     * Liefert die aktuelle Transformationsmatrix.
     * @return Die aktuelle Transformationsmatrix
     */
    public Matrix4f topMat() {
        return this.matStack.get(this.current);
    }
        
    /**
     * Liefert die aktuelle inverse Transformationsmatrix.
     * @return Die aktuelle inverse Transformationsmatrix
     */
    public Matrix4f topMatInverse() {
        return this.matInverseStack.get(this.current);
    }
    
    /**
     * Loescht den gesamten Matrixstack. <i>Hinweis:</i> Ein leerer Matrixstack
     * besitzt immer noch die Einheitsmatrix als Topelement.
     */
    public void reset() {
        this.current = 0;
    }
    
    /**
     * Bestimmt, ob der Stack leer ist. <i>Hinweis:</i> Ein leerer Matrixstack
     * besitzt immer noch die Einheitsmatrix als Topelement.
     * @return <tt>true</tt>, falls leer
     */
    public boolean isEmpty() {
        return this.current == 0;
    }
}
