/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package view.scene;

import actor.Actor;
import actor.components.TransformComponent;
import event.EventData;
import event.EventListener;
import event.events.ActorEvent;
import event.events.ActorMovedEvent;
import graphics.shader.UniformBuffer;
import java.nio.FloatBuffer;
import logging.Logger;
import gameapp.GameApp;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;
import util.MathUtil;
import util.Pose;

/**
 * Klasse fuer eine Perspektivische Kamera.
 * @author nico3000
 */
public class PerspectiveCamera implements EventListener {
    /**
     * Die Pose der TransformComponent zum schnelle Zugriff.
     */
    private Pose pose;
    /**
     * Die ID des zugehoerigen Actors.
     */
    private int actorId;
    /**
     * Perspektivische Projektionsmatrix.
     */
    private final Matrix4f projection = new Matrix4f();
    /**
     * Das Inverse der perspektivischen Projektionsmatrix.
     */
    private final Matrix4f projectionInv = Matrix4f.setZero(new Matrix4f());
    /**
     * <tt>projection * view</tt>.
     */
    private final Matrix4f projectionView = Matrix4f.setIdentity(new Matrix4f());
    /**
     * <tt>viewInv * projectionInv</tt>.
     */
    private final Matrix4f projectionViewInv = Matrix4f.setZero(new Matrix4f());
    /**
     * Sichtweite / Position der entfernten Clipplane (zfar).
     */
    private float viewDistance;
    /**
     * Buffer, der an die GPU geschickt wird.
     */
    private final FloatBuffer bufferData = BufferUtils.createFloatBuffer(52);
    /**
     * Buffer auf der GPU.
     */
    private final UniformBuffer buffer = new UniformBuffer();
    /**
     * Position der nahen Clipplane.
     */
    private float znear = 1e-1f;
    /**
     * Seitenverhältnis der Projektion (Bildschirmbreite / Bildschirmhoehe).
     */
    private float aspectRatio = 4.0f / 3.0f;
    /**
     * Oeffnungswinkel / Field of view der Projektion.
     */
    private float fov = MathUtil.PI_DIV2;
    private Vector4f frustumFaces[] = new Vector4f[] {
        new Vector4f(),
        new Vector4f(),
        new Vector4f(),
        new Vector4f(),
        new Vector4f(),
        new Vector4f(),
    };
    
    /**
     * Initialisiert die Kamera.
     * @param actorId Die ID des mit dieser Kamera vebundenen Actors
     * @param initialFov Das initialte Field of view
     * @param initialAspectRatio Das initiale Seitenverhältnis
     * @param znear Die Entfernung der nahen Clipplane
     * @param zfar Die Entfernung der entfernten Clipplane
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init(Actor actor, float initialFov, float initialAspectRatio, float znear, float zfar) {
        if(this.buffer.init()) {
            if(actor != null) {
                TransformComponent comp = (TransformComponent)actor.getComponent(TransformComponent.ID);
                if(comp == null) {
                    Logger.INSTANCE.error("Actor for PerspectiveCamera without TransformComponent!");
                    return false;
                }
                this.pose = comp.getPose();
                this.actorId = actor.getId();
            } else {
                Logger.INSTANCE.error("Actor for PerspectiveCamera is invalid!");
            }
            this.buffer.setBindingIndex(UniformBuffer.BINDING_CAMERA);
            this.fov = initialFov;
            this.aspectRatio = initialAspectRatio;
            this.znear = znear;
            this.viewDistance = zfar;
            this.updateProjection();
            GameApp.getEventManager().register(this, ActorMovedEvent.ID);
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Zerstoert den Buffer der Kamera.
     */
    public void destroy() {
        this.buffer.destroy();
    }
    
    /**
     * Aktualisiert die Projektionsmatrix. Wird automatisch im Anschluss von
     * <tt>setProjectionData()</tt> aufgerufen.
     */
    private void updateProjection() {
        MathUtil.perspectiveFov(this.projection, this.fov, this.aspectRatio, this.znear, this.viewDistance);
        // inverse of projection
        
        this.projectionInv.m00 = 1.0f / this.projection.m00;
        this.projectionInv.m11 = 1.0f / this.projection.m11;
        this.projectionInv.m32 = 1.0f;
        this.projectionInv.m23 = 1.0f / this.projection.m32;
        this.projectionInv.m33 = -this.projection.m22 / this.projection.m32;
    }

    /**
     * Bindet den Buffer an die GLSL Variable CameraBlock.
     */
    public void bindBuffer() {
        this.buffer.bind();
    }
    
    /**
     * Aktualisiert den Buffer auf der Grafikkarte. Sollte aufgerufen werden,
     * nachdem alle Aenderungen mittels <tt>move()</tt>, <tt>rotate()</tt>, 
     * <tt>updateView()</tt> und <tt>setProjectionData()</tt> vorgenommen
     * wurden.
     */
    public void updateBuffer() {
        Matrix4f.mul(this.projection, this.pose.getModelInv(), this.projectionView);
        Matrix4f.mul(this.pose.getModel(), this.projectionInv, this.projectionViewInv);
        this.bufferData.position(0);
        this.pose.getModelInv().store(this.bufferData);
        this.projection.store(this.bufferData);
        this.projectionView.store(this.bufferData);
        this.pose.getPosition().store(this.bufferData);
        this.bufferData.put(this.viewDistance);
        this.bufferData.position(0);
        this.buffer.setData(this.bufferData);
        this.updateFrustumPlanes();
    }

    /**
     * Setzt die Daten der Projektion.
     * @param fov Das neue Field of view
     * @param aspectRatio Das neue Seitenverhältnis
     */
    public void setProjectionData(float fov, float aspectRatio) {
        this.fov = fov;
        this.aspectRatio = aspectRatio;
        this.updateProjection();
    }

    /**
     * Liefert die Planes, die die Seiten des Frustums definieren.
     * @return Die 6 Seiten des Frustums als Plane
     */
    public Vector4f[] getFrustumFaces() {
        return this.frustumFaces;
    }

    @Override
    public void trigger(EventData data) {
        ActorEvent evt = (ActorEvent)data;
        if(evt.getActorId() == this.pose.getActorId()) {
            this.updateBuffer();
        }
    }
    
    public Pose getPose() {
        return this.pose;
    }

    public int getActorId() {
        return actorId;
    }
    
    private void updateFrustumPlanes() {
        Vector3f v0 = new Vector3f(-1.0f, -1.0f, -1.0f);
        Vector3f v1 = new Vector3f(+1.0f, -1.0f, -1.0f);
        Vector3f v2 = new Vector3f(+1.0f, +1.0f, -1.0f);
        Vector3f v3 = new Vector3f(-1.0f, +1.0f, -1.0f);
        Vector3f v4 = new Vector3f(-1.0f, -1.0f, +1.0f);
        Vector3f v5 = new Vector3f(+1.0f, -1.0f, +1.0f);
        Vector3f v6 = new Vector3f(+1.0f, +1.0f, +1.0f);
        Vector3f v7 = new Vector3f(-1.0f, +1.0f, +1.0f);
        MathUtil.transformCoord(v0, this.projectionViewInv, v0);
        MathUtil.transformCoord(v1, this.projectionViewInv, v1);
        MathUtil.transformCoord(v2, this.projectionViewInv, v2);
        MathUtil.transformCoord(v3, this.projectionViewInv, v3);
        MathUtil.transformCoord(v4, this.projectionViewInv, v4);
        MathUtil.transformCoord(v5, this.projectionViewInv, v5);
        MathUtil.transformCoord(v6, this.projectionViewInv, v6);
        MathUtil.transformCoord(v7, this.projectionViewInv, v7);
        MathUtil.createPlane(this.frustumFaces[0], v4, v0, v1); // -y
        MathUtil.createPlane(this.frustumFaces[1], v7, v6, v2); // +y
        MathUtil.createPlane(this.frustumFaces[2], v1, v2, v6); // +x
        MathUtil.createPlane(this.frustumFaces[3], v0, v4, v7); // -x
        MathUtil.createPlane(this.frustumFaces[4], v0, v3, v2); // -z
        MathUtil.createPlane(this.frustumFaces[5], v5, v6, v7); // +z
        
//        System.out.println("-------------------");
//        for(Vector4f plane : frustumFaces) {
//            System.out.println(plane);
//        }
    }
}
